from schema import Optional, Or, Use, Schema


config_schema = Schema(
    {
        'files': [
            {
                'file': Use(str, error="No file supplied"),
                Optional('direction'): Or('forward', 'reverse'),
                Optional('order_by'): Or('none', 'weight'),
                Optional('rows'): int,
                'patterns': [
                    {
                        'pattern': Use(str),
                        'type': Or('string', 'regex'),
                        Optional('weight'): int,
                        Optional('operator'): Or('AND', 'OR', 'NOT', 'KEYWORD'),
                    }
                ],
                'output': {
                    Optional('path'): Use(str, error="No output path specified"),
                    'type': Or('json', 'html', 'console', error="No output type specified"),
                    Optional('overwrite'): Use(bool)
                },
            }
        ]
    }
)



from argparse import ArgumentParser
from jinja2 import Environment, FileSystemLoader

from filur.models import File
from filur.schemas import config_schema

import yaml
import json
import os


def load_configuration(path: str) -> dict:
    configuration = {}

    if not os.path.exists:
        raise OSError(f"{path} not found")

    try:
        with open(path, 'r', encoding='utf-8') as file:
            configuration = yaml.safe_load(file)

    except yaml.parser.ParserError as err:
        print(f"ParserError {err}")
        return configuration

    else:
        if config_schema.validate(configuration) and configuration is not None:
            return configuration
        return {}


def write_file(path: str, content: str, mode: str = 'w', encoding: str = 'utf-8') -> None:
    try:
        with open(path, mode, encoding=encoding) as file:
            file.write(content)

        print(clickable_link(path, f'Click to open output'))

    except FileNotFoundError:
        print("Error: The directory or file path does not exist.")

    except PermissionError:
        print("Error: Insufficient permissions to write to the file.")

    except IsADirectoryError:
        print("Error: A directory was provided where a file was expected.")

    except OSError as e:
        print(f"OS error occurred: {e}")

    except ValueError:
        print("Error: Invalid file mode or other argument.")


def export_json(configuration: dict, data: dict) -> None:
    output = configuration['output']
    path = output['path']
    overwrite = output['overwrite'] if 'overwrite' in output else False

    if os.path.exists(path) and not overwrite:
        raise OSError(f"File {path} already exists")

    write_file(path, json.dumps(data))


def export_html(configuration: dict, data: dict) -> None:
    output = configuration['output']
    title = configuration['file']
    path = output['path']
    overwrite = output['overwrite'] if 'overwrite' in output else False

    if os.path.exists(path) and not overwrite:
        raise OSError(f"File {path} already exists")

    environment = Environment(loader=FileSystemLoader('templates/'))
    template = environment.get_template('template.html')
    content = template.render(title=title, data=data)

    write_file(path, content)


def clickable_link(uri: str, label: str = '') -> str:
    label = uri if not label else label
    # OSC 8 ; params ; URI ST <name> OSC 8 ;; ST
    return f"\033]8;'';{uri}\033\\{label}\033]8;;\033\\"


def export_output(configuration: dict, data: dict):
    match configuration['output']['type']:
        case 'json':
            export_json(configuration, data)
        case 'html':
            export_html(configuration, data)
        case 'console' | _:
            print(json.dumps(data, indent=2))


def main():
    parser = ArgumentParser()
    parser.add_argument(
        '-p',
        '--playbook',
        required=True,
        help='File path to playbook (yaml) file'
    )
    args = parser.parse_args()
    configuration = load_configuration(args.playbook)

    for file in configuration['files']:
        data = File.from_dict(file)
        processed = data.process()
        export_output(file, processed)


if __name__ == "__main__":
    main()



from .pattern import Pattern

import os
import re


class File:
    def __init__(self, path: str, direction: str, order_by: str, rows: int, output_type: str, output_path: str, patterns: list):
        self.path = path
        self.direction = direction
        self.order_by = order_by
        self.rows = rows
        self.output_type = output_type
        self.output_path = output_path
        self.patterns = patterns

    def _reverse_readline(self, path: str, buf_size: int = 8192):
        """A generator that returns the lines of a file in reverse order"""
        with open(path, 'rb') as fh:
            segment = None
            offset = 0
            fh.seek(0, os.SEEK_END)
            file_size = remaining_size = fh.tell()

            while remaining_size > 0:
                offset = min(file_size, offset + buf_size)
                fh.seek(file_size - offset)
                buffer = fh.read(min(remaining_size, buf_size))

                # remove file's last "\n" if it exists, only for the first buffer
                if remaining_size == file_size and buffer[-1] == ord('\n'):
                    buffer = buffer[:-1]

                remaining_size -= buf_size
                lines = buffer.split('\n'.encode())

                # append last chunk's segment to this chunk's last line
                if segment is not None:
                    lines[-1] += segment

                segment = lines[0]
                lines = lines[1:]

                # yield lines in this chunk except the segment
                for line in reversed(lines):
                    # only decode on a parsed line, to avoid utf-8 decode error
                    yield line.decode()

            # Don't yield None if the file was empty
            if segment is not None:
                yield segment.decode()

    def _read_file(self) -> list:
        data = []

        if self.direction == 'reverse':
            data = self._reverse_readline(self.path)

        else:
            with open(self.path, 'r', encoding='utf-8') as file:
                data = file.readlines()

        return data

    def _matches(self, patterns: list, row: str) -> dict:
        result = {
            'patterns': [],
            'matches': []
        }

        for pattern in patterns:
            if self._matches_pattern(row, pattern):
                result['patterns'].append(pattern)
                result['matches'].append(True)
            else:
                result['matches'].append(False)

        return result

    def _match_row(self, row: str) -> dict:
        result = {}
        matched_patterns = []
        and_conditions = [
            pattern for pattern in self.patterns if pattern.operator.upper() == 'AND']
        or_conditions = [
            pattern for pattern in self.patterns if pattern.operator.upper() == 'OR']
        not_conditions = [
            pattern for pattern in self.patterns if pattern.operator.upper() == 'NOT']
        keyword_conditions = [
            pattern for pattern in self.patterns if pattern.operator.upper() == 'KEYWORD']
        and_matches = self._matches(and_conditions, row)
        or_matches = self._matches(or_conditions, row)
        not_matches = self._matches(not_conditions, row)
        keyword_matches = self._matches(keyword_conditions, row) if and_conditions or or_conditions else {}
        and_match = all(and_matches['matches'])
        or_match = any(or_matches['matches'])
        not_match = all(not_matches['matches'])

        for match in [and_matches, or_matches, not_matches, keyword_matches]:
            for pattern in match['patterns']:
                matched_patterns.append(pattern.to_dict())

        if (not and_conditions or and_match) and (not or_conditions or or_match) and (not not_conditions or not not_match):
            result['patterns'] = matched_patterns
            result['combined_weight'] = sum(
                [pattern['weight'] for pattern in matched_patterns])
            result['row'] = row

        return result

    def _matches_pattern(self, row: str, pattern) -> bool:
        match pattern.type:
            case 'regex':
                return re.search(pattern.pattern, row) is not None
            case 'string':
                return pattern.pattern in row
            case _:
                return False

    def _match_rows(self, data: list) -> dict:
        result = {
            'matched_rows': 0,
            'rows': []
        }

        for idx, row in enumerate(data):
            if not self.rows == -1 and idx >= self.rows:
                break

            row = self._match_row(row)

            if row:
                result['rows'].append(row)
                result['matched_rows'] += 1

        if self.order_by == 'weight':
            result['rows'] = sorted(
                result['rows'], key=lambda d: d['combined_weight'], reverse=True)

        return result

    def process(self):
        if not os.path.exists(self.path):
            raise OSError(f"{self.path} not found")

        data = self._read_file()
        result = self._match_rows(data)

        return result

    @classmethod
    def from_dict(cls, data: dict):
        # Set defaults in case nothing is specified in the playbook
        direction = data.get('direction', 'forward')
        order_by = data.get('order_by', 'none')
        rows = data.get('rows', -1)
        output = data.get('output', {})
        output_type = output.get('type', 'console')
        output_path = output.get('path', '')

        # Create Pattern objects from each pattern
        patterns = [Pattern(**pattern_data)
                    for pattern_data in data.get('patterns', [])]

        # Instantiate and return a File object with the Pattern objects
        return cls(
            path=data['file'],
            direction=direction,
            order_by=order_by,
            rows=rows,
            output_type=output_type,
            output_path=output_path,
            patterns=patterns
        )

    def __repr__(self):
        return f"File(path={self.path}, direction={self.direction}, order_by={self.order_by}, rows={self.rows}, output_type={self.output_type}, output_path={self.output_path}, patterns={self.patterns})"



class Pattern:
    def __init__(self, pattern: str, type: str = 'string', weight: int = 1, operator: str = 'OR'):
        self.pattern = pattern
        self.type = type
        self.weight = weight
        self.operator = operator

    def to_dict(self):
        return {
            'pattern': self.pattern,
            'type': self.type,
            'weight': self.weight,
            'operator': self.operator
        }

    def __repr__(self):
        return f"Pattern(pattern={self.pattern}, type={self.type}, weight={self.weight}, operator={self.operator})"
